##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

r"""
PyLipID class
=============
**api** provides the outer layer functions to calculate lipid interactions
and binding sites.

.. currentmodule:: pylipid.api

.. autosummary::
   :toctree: generated/

   LipidInteraction

"""

from .api import LipidInteraction